#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "lua.h"
#include "lualib.h"
#include "luadebug.h"
#include "ldblib.h"

static void verify_line_counters (void);
static void execute_global_action (void);
static void check_current_location (void);
static void call_procedure (char *file, int line);
static void verify_call_stops (void);
static void verify_call_counters (void);
static void return_procedure (char *file, int line);
static void verify_return_stops (void);
static void command_line (void);
static void ldb_linehook (int line);
static void ldb_callhook (lua_Object func, char *file, int line);
static void execute_command (char *s);
static int local_env (int level);
static int global_env (int level);
static void list_lines (char *file_name, int first, int num_lines);
static void ldb_next (void);

#define IN_RUN_MODE	(mode == RUN_MODE)
#define IN_STEP_MODE	(mode == STEP_MODE)
#define IN_NEXT_MODE	(mode == NEXT_MODE)
#define TAM_BUF 512
#define ldbI_existbreak()	(last_mark > 0)
#define ldbI_exprfunc(_)	(exprs[_])
#define ldbI_existGAction()	(GActionString)
#define ldbI_GAction()	(GActionFunc)
#define ldbI_GActionString()	(GActionString)

typedef struct {
  int line;
  char *file;
} location;
typedef struct Marks
{
  location loc;
  int  code;
} Marks;
#define MAX_EXPRESSIONS 30
static int exprs[MAX_EXPRESSIONS];

static location curr_loc = {0, NULL};
static int stop_now = 0;
static int line_count;
enum {RUN_MODE, STEP_MODE, NEXT_MODE} mode = RUN_MODE;
enum {OFF_HOOK, ON_HOOK, DONT_TURN_ON} line_on = OFF_HOOK,
                                       call_on = OFF_HOOK;
static int func_level;
static int shortcut[256];
static int global_ref;
static int curr_level = 0;
#define MAX_MARKS 30
static Marks marks[MAX_MARKS];
static int last_mark;
static int last_expr;
static char *GActionString = NULL;
static int GActionFunc = -1;

static int create_action (char *body, char *params)
{
  char s[512];
  int i;
  lua_beginblock ();
    sprintf (s, "function dummy (%s) %s end", params, body);
    if (lua_dostring(s))
      lua_error ("ldb: incorrect action format");
    lua_pushobject (lua_getglobal ("dummy"));
    i = lua_lock ();
  lua_endblock ();
  return i;
}
static void do_action (int action, char *arg)
{
  lua_Object func;
  lua_beginblock ();
    func = lua_getlocked (action);
    if (arg)
      lua_pushstring (arg);
    lua_callfunction (func);
  lua_endblock ();
}
static void ldbI_initmarks (void)
{
  last_mark = 0;
}
static int ldbI_alreadyset (location *loc)
{
  int i = 0;
  while (i < last_mark)
  {
    if (marks[i].loc.line == loc->line)
      if (!strcmp (marks[i].loc.file, loc->file))
        return i;
    i++;
  }
  return last_mark;
}
static void ldbI_setbreak (location *loc, int code)
{
  int m = ldbI_alreadyset (loc);
  if (m == last_mark)
    if (last_mark < MAX_MARKS)
      last_mark++;
    else
    {
      /* too much break points */
      return;
    }
  marks[m].loc = *loc;
  marks[m].code = code;
}
static int ldbI_getcode (location *loc)
{
  int i = ldbI_alreadyset (loc);
  if (i == last_mark)
    return 0;
  else
    return marks[i].code;
}
static void ldbI_delete (location *loc)
{
  int i = 0;
  while (i < last_mark)
  {
    if (marks[i].loc.line == loc->line)
      if (!strcmp (marks[i].loc.file, loc->file))
      {
        lua_unlock (marks[i].code);
        last_mark--;
        marks[i] = marks[last_mark];
        return;
      }
    i++;
  }
}
static int ldbI_getbreak (int n, location *loc, int *code)
{
  if (n >= last_mark)
    return 0;
  *loc = marks[n].loc;
  *code = marks[n].code;
  return 1;
}
static void ldbI_initexprs (void)
{
  int i;
  for (i=0; i<MAX_EXPRESSIONS; i++)
    exprs[i] = 0;
  last_expr = 0;
}
static void ldbI_setdisplay (int func_ref)
{
  int i;
  for (i=0; i<last_expr; i++)
    if (exprs[i] == 0)
    {
      exprs[i] = func_ref;
      return;
    }
  if (last_expr == MAX_EXPRESSIONS)
    lua_error ("ldb: too many expressions to display");
  exprs[last_expr] = func_ref;
  last_expr++;
}
static void ldbI_unsetdisplay (int exp_number)
{
  if ((exp_number-1) < last_expr && exp_number > 0)
  {
    exprs[exp_number-1] = 0;
    if (last_expr == exp_number-1)
      while (exprs[last_expr-1] == 0)
        last_expr--;
  }
}
static void ldbI_setGAction (char *expr)
{
  if (ldbI_existGAction ())
  {
    free (GActionString);
    lua_unlock (GActionFunc);
  }
  if (expr)
  {
    GActionFunc = create_action (expr, "");
    GActionString = strdup (expr);
  }
  else
  {
    GActionString = NULL;
    GActionFunc = -1;
  }
}

static void update_location (int level, int line)
{
  lua_Object f = lua_stackedfunction (level);
  lua_funcinfo (f, &curr_loc.file, &curr_loc.line);
  if ((line != -1)  || ((line = lua_currentline (f)) != -1))
    curr_loc.line = line;
}
static void disable_hooks (void)
{
  if (lua_callhook)
  {
    lua_callhook = NULL;
    call_on = ON_HOOK;
  }
  else
    call_on = DONT_TURN_ON;
  if (lua_linehook)
  {
    lua_linehook = NULL;
    line_on = ON_HOOK;
  }
  else
    line_on = DONT_TURN_ON;
}
static void restore_hooks (void)
{
  if (call_on == ON_HOOK)
    lua_callhook = ldb_callhook;
  call_on = OFF_HOOK;
  if (line_on == ON_HOOK)
    lua_linehook = ldb_linehook;
  line_on = OFF_HOOK;
}
static void enable_line_hook (void)
{
  if (line_on == OFF_HOOK)
    lua_linehook = ldb_linehook;
  else if (line_on == DONT_TURN_ON)
    line_on = ON_HOOK;
}
static void enable_call_hook (void)
{
  if (call_on == OFF_HOOK)
    lua_callhook = ldb_callhook;
  else if (call_on == DONT_TURN_ON)
    call_on = ON_HOOK;
}
static void maybe_disable_hooks (void)
{
  if (!ldbI_existbreak() && !ldbI_existGAction() && IN_RUN_MODE)
  {
    lua_linehook = NULL;
    if (line_on == ON_HOOK)
      line_on = DONT_TURN_ON;
    lua_callhook = NULL;
    if (call_on == ON_HOOK)
      call_on = DONT_TURN_ON;
  }
}
static void verify_line_stops (void)
{
  execute_global_action ();
  check_current_location ();
  verify_line_counters ();
}
static void verify_line_counters (void)
{
  if (IN_STEP_MODE || (IN_NEXT_MODE && !func_level))
  {
    line_count--;
    if (line_count == 0)
    {
      mode = RUN_MODE;
      stop_now = 1;
    }
  }
}
static void execute_global_action (void)
{
  if (ldbI_existGAction ())
  {
    disable_hooks ();
    do_action (ldbI_GAction (), NULL);
    restore_hooks ();
  }
}
static void check_current_location (void)
{
  int code = ldbI_getcode (&curr_loc);
  if (code < 0)
    stop_now = 1;
  else if (code > 0)
  {
    disable_hooks ();
    local_env (0);
    do_action (code, NULL);
    global_env (0);
    restore_hooks ();
  }
}
static void call_procedure (char *file, int line)
{
  curr_loc.file = file;
  curr_loc.line = line;
  verify_call_stops ();
  if (stop_now)
    command_line ();
}
static void verify_call_stops (void)
{
  execute_global_action ();
  check_current_location ();
  verify_call_counters ();
}
static void verify_call_counters (void)
{
  if (IN_NEXT_MODE)
    func_level++;
  else if (IN_STEP_MODE)
  {
    line_count--;
    if (line_count == 0)
    {
      mode = RUN_MODE;
      stop_now = 1;
    }
  }
}
static void return_procedure (char *file, int line)
{
  verify_return_stops ();
  if (stop_now)
  {
    curr_loc.line = 0;
    curr_loc.file = "<return>";
    command_line ();
  }
  update_location (1, -1);
}
static void verify_return_stops (void)
{
  if (IN_NEXT_MODE)
  {
    if (--func_level < 0)
    {
      mode = RUN_MODE;
      stop_now = 1;
    }
  }
  else if (IN_STEP_MODE)
  {
    line_count--;
    if (line_count == 0)
    {
      mode = RUN_MODE;
      stop_now = 1;
    }
  }
}
static void show_expressions (void)
{
  int i;
  for (i=0; i<last_expr; i++)
  {
    lua_beginblock ();
      if (ldbI_exprfunc (i) != 0)
      {
        char s[256];
        lua_Object expr = lua_getlocked (ldbI_exprfunc(i));
        char *e = lua_getstring (expr);
        fprintf (stderr, "%d: %s = ", i+1, e);
        sprintf (s, "print(%s)", e);
        lua_dostring (s);
      }
    lua_endblock ();
  }
}
static void command_line (void)
{
  char s[256];
  disable_hooks ();
  curr_level = 0;
  local_env (0);
  mode = RUN_MODE;
  show_expressions ();
  if (curr_loc.line <= 0)
  fprintf (stderr, "<%s>\n", curr_loc.file);
else
  list_lines (curr_loc.file, curr_loc.line, 1);
fprintf (stderr, "ldb>");

  while (gets (s) != 0)
  {
    execute_command (s);
    if (!stop_now)
      break;
    fprintf (stderr, "ldb>");
  }
  global_env (curr_level);
  restore_hooks ();
}
static void execute_command (char *s)
{
  static char old_command = 0;
  if (s[0] == 0)
  {
    s[0] = '\'';
    s[1] = old_command;
  }
  if (s[0] != '\'')
    lua_dostring (s);
  else
  {
    if (s[1])
    {
      int action = shortcut[(int)s[1]];
      old_command = s[1];
      do_action (action, s+2);
    }
  }
}
static void setshortcut (int k, char *s)
{
  shortcut[k] = create_action (s, "line");
}
static void shortcut_init (void)
{
  int i;
  int action = create_action ("error('Shortcut not defined')", "");
  for (i=0; i < 256; i++)
    shortcut[i] = action;
}
static int local_env (int level)
{
  char       *name;
  lua_Object func = lua_stackedfunction (level);
  lua_Object table;
  lua_Object local;
  int        n = 1;
  if (func == LUA_NOOBJECT)
    return 0;
  else if (lua_type (func) == LUA_T_CMARK)
    return -1;
  lua_beginblock ();
    table = lua_getlocked (global_ref);
    while ((local = lua_getlocal (func, n, &name)) != LUA_NOOBJECT)
    {
      lua_Object value;
      lua_beginblock ();
        value = lua_getglobal (name);
if (!lua_isnil (value))
{
  lua_pushobject (table);
  lua_pushstring (name);
  lua_pushobject (value);
  lua_storesubscript ();
}

        lua_pushobject (local);
lua_storeglobal (name);

      lua_endblock ();
      n++;
    }
  lua_endblock ();
  return 1;
}
static int global_env (int level)
{
  char       *name;
  lua_Object func = lua_stackedfunction (level);
  lua_Object table;
  int        n = 1;
  if (func == LUA_NOOBJECT)
    return 0;
  else if (lua_type (func) == LUA_T_CMARK)
    return -1;
  lua_beginblock ();
    table = lua_getlocked (global_ref);
    while (lua_getlocal (func, n, &name) != LUA_NOOBJECT)
    {
      lua_Object global;
      lua_beginblock ();
        lua_pushobject (lua_getglobal (name));
lua_setlocal (func, n);

        /* global = global_ref[name] */
lua_pushobject (table);
lua_pushstring (name);
global = lua_getsubscript ();
/* name = global */
lua_pushobject (global);
lua_storeglobal (name);
/* global_ref[name] = nil */
lua_pushobject (table);
lua_pushstring (name);
lua_pushnil ();
lua_storesubscript ();

      lua_endblock ();
      n++;
    }
  lua_endblock ();
  return 1;
}
static int get_location (int p1, location *loc)
{
  lua_Object o1 = lua_getparam (p1);
  lua_Object o2 = lua_getparam (p1+1);
  if (lua_isfunction (o1))
    lua_funcinfo (o1, &(loc->file), &(loc->line));
  else if (lua_isnumber (o1))
  {
    loc->line = (int)lua_getnumber (o1);
    if (lua_isstring (o2))
      loc->file = lua_getstring (o2);
    else
      loc->file = curr_loc.file;
  }
  else
    lua_error ("Invalid identification of source point");
  return lua_type (o1);
}
static void enable_hooks (int is_function)
{
  if (is_function)
    enable_call_hook ();
  else
    enable_line_hook ();
}
static void list_lines (char *file_name, int first_line, int num_lines)
{
  FILE *arq;
  int i;
  char buffer[TAM_BUF+1];
  if ((arq = fopen(file_name, "r")) == NULL)
    lua_error ("can't open file");
  for (i=1; i<first_line; i++)
    fgets (buffer, TAM_BUF, arq);
  for (i=0; i<num_lines; i++)
  {
    if (fgets (buffer, TAM_BUF, arq) == NULL)
      break;
    fprintf (stderr, "%d\t%s", first_line+i, buffer);
  }
  fclose (arq);
}
static void ldb_error (void)
{
  lua_Object o = lua_getparam(1);
  if (lua_isstring (o))
    fprintf (stderr, "lua: %s\n", lua_getstring(o));
  else
    fprintf(stderr, "lua: unknown error\n");
  enable_line_hook ();
  enable_call_hook ();
  stop_now = 1;
}
static void arg_error(char *funcname)
{
  char buff[100];
  sprintf (buff, "incorrect arguments to function `%s'", funcname);
  lua_error (buff);
}
static char *check_string (int numArg, char *funcname)
{
  lua_Object o = lua_getparam(numArg);
  if (!lua_isstring(o))
    arg_error(funcname);
  return lua_getstring (o);
}
static int check_number (int numArg, char *funcname)
{
  lua_Object o = lua_getparam(numArg);
  if (!lua_isnumber(o))
    arg_error(funcname);
  return (int)lua_getnumber(o);
}
static int get_opt_number (int numArg, int def, char *funcname)
{
  return (lua_getparam(numArg) == LUA_NOOBJECT) ? def :
                              (int)check_number(numArg, funcname);
}
void load_config_file (void)
{
  char *config_file_name = getenv("LDB_CONFIG");
  if (config_file_name == NULL)
    config_file_name = "ldbrc.lua";
  if (lua_dofile(config_file_name) != 0)
    fprintf (stderr, "Warning: Lua debugger cannot load its configuration file"
             " (%s)\n", config_file_name);
}

static void ldb_break (void)
{
  location loc;
  if (get_location (1, &loc) == LUA_T_NUMBER)
    enable_line_hook ();
  else
    enable_call_hook ();
  ldbI_setbreak (&loc, -1);
}
static void ldb_cont (void)
{
  stop_now = 0;
  maybe_disable_hooks ();
}
static void ldb_delete (void)
{
  location loc;
  get_location (1, &loc);
  ldbI_delete (&loc);
  maybe_disable_hooks ();
}
static void ldb_depth (void)
{
  int level = get_opt_number(1, curr_level, "depth");
  global_env (curr_level+2);
  if (local_env (level+2))
    curr_level = level;
  else
    local_env (curr_level+2);
  lua_pushnumber (curr_level);
}
static void ldb_display (void)
{
  int value;
  char *s = check_string (1, "display");
  lua_pushstring (s);
  value = lua_lock ();
  ldbI_setdisplay (value);
}
static void ldb_doalways (void)
{
  lua_Object o = lua_getparam (1);
  if (o == LUA_NOOBJECT)
    ldbI_setGAction (NULL);
  else
    ldbI_setGAction (check_string (1, "doalways"));
  enable_line_hook ();
  enable_call_hook ();
}
static void ldb_doat (void)
{
  int value;
  location loc;
  if (get_location (2, &loc) == LUA_T_NUMBER)
    enable_line_hook ();
  else
    enable_call_hook ();
  value = create_action (check_string(1, "doat"), "");
  ldbI_setbreak (&loc, value);
}
static void ldb_down (void)
{
  int level = get_opt_number(1, 1, "down");
  global_env (curr_level+2);
  if (local_env (curr_level+2+level))
    curr_level += level;
  else
    local_env (curr_level+2);
  enable_call_hook ();
}
static void ldb_file (void)
{
  if (curr_loc.file)
    lua_pushstring (curr_loc.file);
  else
    lua_pushnil ();
}
static void ldb_finish (void)
{
  line_count = -1;
  mode = NEXT_MODE;
  func_level = 0;
  stop_now = 0;
  enable_line_hook ();
  enable_call_hook ();
/*
  lua_pushnumber (-1);
  ldb_next ();
*/
}
static void ldb_goto (void)
{
  int value;
  location loc;
  int is_function = (get_location (1, &loc) != LUA_T_NUMBER);
  {
  char s[TAM_BUF];
  sprintf (s, "delete(%d,'%s') stop()", loc.line, loc.file);
  value = create_action (s, "");
}

  ldbI_setbreak (&loc, value);
  stop_now = 0;
  enable_hooks (is_function);
}
static void ldb_line (void)
{
  lua_pushnumber (curr_loc.line);
}
static void ldb_list (void)
{
  lua_Object o1 = lua_getparam (1);
  lua_Object o2 = lua_getparam (2);
  lua_Object o3 = lua_getparam (3);
  unsigned num_lines = 0;		/* to avoid warnings */
  int first_line;
  char *file_name;
  if (lua_isnumber(o1))
{
  num_lines = lua_getnumber(o1);
  if (lua_isfunction(o2))
    lua_funcinfo (o2, &file_name, &first_line);  
  else if (lua_isstring (o2) && !lua_isnumber(o2))
  {
    first_line = 1;
    file_name = lua_getstring(o2);
  }
  else
  {
    first_line = lua_isnumber(o2) ? lua_getnumber(o2) :
	(curr_loc.line==0 ? 1 : curr_loc.line);
    file_name = lua_isstring(o3) ? lua_getstring (o3) : curr_loc.file;
  }
}

else if (lua_isstring(o1))
{
  num_lines = -1;
  first_line = 1;
  file_name = lua_getstring(o1);
}
else
  lua_error ("first parameter must be a number");

  list_lines(file_name, first_line, num_lines);
}
static void ldb_next (void)
{
  line_count = get_opt_number (1, 1, "Next");
  mode = NEXT_MODE;
  func_level = 0;
  stop_now = 0;
  enable_line_hook ();
  enable_call_hook ();
}
static void ldb_setshortcut (void)
{
  char *s1 = check_string(1, "setshortcut");
  char *s2 = check_string(2, "setshortcut");
  setshortcut(*s1, s2);
}
static void ldb_show (void)
{
  int code, i;
  location loc;
  for (i=0; ldbI_getbreak (i, &loc, &code); i++)
    fprintf (stderr, "[%d:%s]\n", loc.line, loc.file);
  if (ldbI_existGAction ())
    fprintf (stderr, ">>%s<<\n", ldbI_GActionString ());
}
static void ldb_step (void)
{
  line_count = get_opt_number (1, 1, "step");
  mode = STEP_MODE;
  stop_now = 0;
  enable_line_hook ();
  enable_call_hook ();
}
static void ldb_stop (void)
{
  lua_Object o = lua_getparam (1);
  if (o != LUA_NOOBJECT && lua_isnil(o))
    return;
  stop_now = 1;
  enable_line_hook ();
  enable_call_hook ();
}
static void ldb_undisplay (void)
{
  ldbI_unsetdisplay (check_number (1, "undisplay"));
}
static void ldb_up (void)
{
  int old_level = curr_level;
  int level = get_opt_number (1, 1, "up");
  global_env (curr_level+2);
  curr_level -= level;
  if (curr_level < 0)
    curr_level = 0;
  if (local_env (curr_level+2) == 0)
  {
    curr_level = old_level;
    local_env (curr_level);
  }
}
static void ldb_where (void)
{
  int first_level = 0;
  lua_Object func;
  lua_Object o = lua_getparam (1);
  if (lua_isnumber (o))
    first_level = (int)lua_getnumber (o);
  fprintf(stderr, "Active Stack:\n");
  while ((func = lua_stackedfunction(first_level++)) != LUA_NOOBJECT)
  {
    char *name;
    int currentline;
    char *filename;
    int linedefined;
    fprintf(stderr, "\t");
    lua_funcinfo(func, &filename, &linedefined);
    if (linedefined == 0)
      fprintf(stderr, "main of %s", filename);
    else
    {
      switch (*lua_getobjname(func, &name))
      {
        case 'g':
          fprintf(stderr, "function %s", name);
          break;
        case 'f':
          fprintf(stderr, "`%s' fallback", name);
          break;
        default:
          fprintf(stderr, "function defined at ");
      }
      if (linedefined < 0)
        fprintf(stderr, " (%s)", filename);
      else
        fprintf(stderr, " (%s:%d)", filename, linedefined);
    }
    if ((currentline = lua_currentline(func)) > 0)
      fprintf(stderr, " - at line %d", currentline);
    fprintf(stderr, "\n");
  }
}

static void ldb_linehook (int line)
{
  if (!curr_loc.file)
    update_location (0, line);
  else
    curr_loc.line = line;
  verify_line_stops ();
  if (stop_now)
    command_line ();
}
static void ldb_callhook (lua_Object func, char *file, int line)
{
  if (func == LUA_NOOBJECT)
    return_procedure (file, line);
  else
    call_procedure (file, line);
}

static struct FuncList {
  char *name;
  lua_CFunction func;
} FuncList [] = {
                {"break", ldb_break},
{"cont", ldb_cont},
{"delete", ldb_delete},
{"depth", ldb_depth},
{"display", ldb_display},
{"doalways", ldb_doalways},
{"doat", ldb_doat},
{"down", ldb_down},
{"file", ldb_file},
{"finish", ldb_finish},
{"goto", ldb_goto},
{"line", ldb_line},
{"list", ldb_list},
{"Next", ldb_next},
{"setshortcut", ldb_setshortcut},
{"show", ldb_show},
{"step", ldb_step},
{"stop", ldb_stop},
{"undisplay", ldb_undisplay},
{"up", ldb_up},
{"where", ldb_where},

                {"stack", ldb_where}
                };

#define SizeFuncList (sizeof(FuncList)/sizeof(struct FuncList))


void ldblib_open (void)
{
  int i;
  static char *revision = "$Revision: 1.6 $\b";
  fprintf (stderr, "LDB %s \n", revision+1);
  ldbI_initmarks ();
  ldbI_initexprs ();
  lua_pushobject (lua_createtable ());
global_ref = lua_lock ();

  for (i = 0; i < SizeFuncList; i++)
    lua_register(FuncList[i].name, FuncList[i].func);
  shortcut_init ();
  load_config_file ();
  lua_setfallback ("error", ldb_error);
}



